/* See float_image.h */
/* Last edited on 2007-11-01 17:40:12 by stolfi */ 

#include <float_image.h>

#include <indexing.h>
#include <filefmt.h>
#include <nget.h>
#include <fget.h>
#include <affirm.h>
#include <bool.h>

#include <values.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* INTERNAL PROTOTYPES */

void float_image_get_interpolation_data
  ( float_image_t *fim,
    int c,
    double x, 
    double y,
    float **p00,
    float **p10,
    float **p01,
    float **p11,
    double *wx0,
    double *wx1,
    double *wy0, 
    double *wy1
  );
  /* Computes the addresses {*p00,*p01,*p10,*p11} of the four samples
    of channel {c} that are involved in the interpolation of image {fim}
    at the point {(x,y)}, and the corresponding weights.

    The first digit in the pointer's name is the X (column) increment,
    the second index is the row increment. Therefor, {*p00} is the
    pixel with lowest X and Y indices among the four; {*p10} is its
    neighbor in the next column of the same row; and so on.

    The pixel addresses are clipped against the number of columns {NX}
    and the number of rows {NY}. Namely, if {(x,y)} lies outside the
    domain rectangle {[0_NX]  [0_NY]}, then the addresses
    {*p00,*p01,*p10,*p11} get adjusted so that they point to valid
    surrogate samples. These adjustment have the same effect as
    extending the image {fim} to infinity, by duplicating its border
    rows and columns. */

void float_image_get_interp_indices_and_weights
  ( double z, 
    int N, 
    int *i0, 
    int *i1, 
    double *w0,
    double *w1
  );
  /* Computes the indices {i0,i1} of the pixels nedded to interpolate
    a row (or column) of samples at the fractional coordinate {z}, and
    their respective weights {w0,w1}. The indices are consecutive,
    except that they are clipped to the range {0..N-1}; therefore
    {i1 - i0} is either 0 or 1. */

/* IMPLEMENTATIONS */

float float_image_get_sample(float_image_t *fim, int c, int x, int y)
  { assert((c >= 0) && (c < fim->sz[0]));
    assert((x >= 0) && (x < fim->sz[1]));
    assert((y >= 0) && (y < fim->sz[2]));
    ix_index_t ix[3] = {c, x, y};
    ix_pos_t p = ix_position(3, ix, fim->bp, fim->st);
    return fim->sample[p];
  }

float *float_image_get_sample_address(float_image_t *fim, int c, int x, int y)
  { assert((c >= 0) && (c < fim->sz[0]));
    assert((x >= 0) && (x < fim->sz[1]));
    assert((y >= 0) && (y < fim->sz[2]));
    ix_index_t ix[3] = {c, x, y};
    ix_pos_t p = ix_position(3, ix, fim->bp, fim->st);
    return &(fim->sample[p]);
  }

void float_image_set_sample(float_image_t *fim, int c, int x, int y, float v)
  { assert((c >= 0) && (c < fim->sz[0]));
    assert((x >= 0) && (x < fim->sz[1]));
    assert((y >= 0) && (y < fim->sz[2]));
    ix_index_t ix[3] = {c, x, y};
    ix_pos_t p = ix_position(3, ix, fim->bp, fim->st);
    fim->sample[p] = v;
  }

void float_image_get_pixel(float_image_t *fim, int x, int y, float v[])
  { float *sp = float_image_get_sample_address(fim, 0, x, y);
    int NC = fim->sz[0];
    int c;
    for (c = 0; c < NC; c++) { v[c] = (*sp); sp += fim->st[0]; }
  }

void float_image_set_pixel(float_image_t *fim, int x, int y, float v[])
  { float *sp = float_image_get_sample_address(fim, 0, x, y);
    int NC = fim->sz[0];
    int c;
    for (c = 0; c < NC; c++) { (*sp) = v[c]; sp += fim->st[0]; }
  }

void float_image_fill_pixel(float_image_t *fim, int x, int y, float v)
  { float *sp = float_image_get_sample_address(fim, 0, x, y);
    int NC = fim->sz[0];
    int c;
    for (c = 0; c < NC; c++) { (*sp) = v; sp += fim->st[0]; }
  }


void float_image_get_sample_row(float_image_t *fim, int c, int y, float v[])
  { float *sp = float_image_get_sample_address(fim, c, 0, y);
    int NX = fim->sz[1];
    int x;
    for (x = 0; x < NX; x++) { v[x] = (*sp); sp += fim->st[1]; }
  }

void float_image_set_sample_row(float_image_t *fim, int c, int y, float v[])
  { float *sp = float_image_get_sample_address(fim, c, 0, y);
    int NX = fim->sz[1];
    int x;
    for (x = 0; x < NX; x++) { (*sp) = v[x]; sp += fim->st[1]; }
  }

void float_image_fill_sample_row(float_image_t *fim, int c, int y, float v)
  { float *sp = float_image_get_sample_address(fim, c, 0, y);
    int NX = fim->sz[1];
    int x;
    for (x = 0; x < NX; x++) { (*sp) = v; sp += fim->st[1]; }
  }


void float_image_get_pixel_row(float_image_t *fim, int y, float v[])
  { float *pxp = float_image_get_sample_address(fim, 0, 0, y);
    int NC = fim->sz[0];
    int NX = fim->sz[1];
    int c, x; int k = 0;
    for (x = 0; x < NX; x++)
      { float *sp = pxp;
        for (c = 0; c < NC; c++) 
          { v[k] = (*sp); sp += fim->st[0]; k++; }
        pxp += fim->st[1];
      }
  }

void float_image_set_pixel_row(float_image_t *fim, int y, float v[])
  { float *pxp = float_image_get_sample_address(fim, 0, 0, y);
    int NC = fim->sz[0];
    int NX = fim->sz[1];
    int c, x; int k = 0;
    for (x = 0; x < NX; x++)
      { float *sp = pxp;
        for (c = 0; c < NC; c++) 
          { (*sp) = v[k]; sp += fim->st[0]; k++; }
        pxp += fim->st[1];
      }
  }

void float_image_fill_pixel_row(float_image_t *fim, int y, float v)
  { float *pxp = float_image_get_sample_address(fim, 0, 0, y);
    int NC = fim->sz[0];
    int NX = fim->sz[1];
    int c, x;
    for (x = 0; x < NX; x++)
      { float *sp = pxp;
        for (c = 0; c < NC; c++) 
          { (*sp) = v; sp += fim->st[0]; }
        pxp += fim->st[1];
      }
  }

void float_image_fill_channel(float_image_t *fim, int c, float v)
  { int NC = fim->sz[0];
    int NX = fim->sz[1]; 
    int NY = fim->sz[2];
    demand((c >= 0) && (c < NC), "invalid channel");
    int x, y;
    for (y = 0; y < NY; y++)
      { for (x = 0; x < NX; x++)
          { float_image_set_sample(fim, c, x, y, v); }
      }
  }

void float_image_fill(float_image_t *fim, float v)
  { int NC = fim->sz[0];
    int c;
    for (c = 0; c < NC; c++) { float_image_fill_channel(fim, c, v); }
  }

/* IMAGE INTERPOLATION */

float float_image_interpolate_samples(float_image_t *fim, int c, double x, double y)
  { /* Get pixel addresses and weights: */
    double wx0, wx1;    /* Column weights. */
    double wy0, wy1;    /* Sample weights. */
    float *p00, *p10, *p01, *p11; /* Sample addresses. */
    float_image_get_interpolation_data
      (fim, c, x, y, &p00, &p10, &p01, &p11, &wx0, &wx1, &wy0, &wy1);
    
    /* Apply interpolation formula: */
    return ((*p00)*wx0 + (*p10)*wx1)*wy0 + ((*p01)*wx0 + (*p11)*wx1)*wy1;
  }
  
void float_image_interpolate_pixels(float_image_t *fim, double x, double y, float v[])
  { /* Get pixel addresses and weights: */
    double wx0, wx1;    /* Column weights. */
    double wy0, wy1;    /* Sample weights. */
    float *p00, *p10, *p01, *p11; /* Sample addresses. */
    float_image_get_interpolation_data
      (fim, 0, x, y, &p00, &p10, &p01, &p11, &wx0, &wx1, &wy0, &wy1);
    
    /* Apply interpolation formula to all channels: */
    int NC = fim->sz[0];         /* Number of channels. */
    ix_step_t cst = fim->st[0];  /* Position increment between channels. */
    int c;
    for (c = 0; c < NC; c++)
      { /* Interpolate: */
        v[c] = ((*p00)*wx0 + (*p10)*wx1)*wy0 + ((*p01)*wx0 + (*p11)*wx1)*wy1;
        /* Advance sample pointers to next channel: */
        p00 += cst; p10 += cst;  p01 += cst;  p11 += cst;
      }
  }

void float_image_get_interpolation_data
  ( float_image_t *fim,
    int c,
    double x, 
    double y,
    float **p00,
    float **p10,
    float **p01,
    float **p11,
    double *wx0,
    double *wx1,
    double *wy0, 
    double *wy1
  )
  {
    int ix0, ix1;       /* Column indices to interpolate. */
    float_image_get_interp_indices_and_weights(x, fim->sz[1], &ix0, &ix1, wx0, wx1);
    int dx = ix1 - ix0; /* Usually 1, may be 0. */
    
    int iy0, iy1;       /* Row indices to interpolate. */
    float_image_get_interp_indices_and_weights(y, fim->sz[2], &iy0, &iy1, wy0, wy1);
    int dy = iy1 - iy0; /* Usually 1, may be 0. */
    
    /* Get sample addresses: */
    (*p00) = float_image_get_sample_address(fim, c, ix0, iy0);
    (*p10) = (*p00) + dx*fim->st[1];
    (*p01) = (*p00) + dy*fim->st[2];
    (*p11) = (*p10) + dy*fim->st[2];
  }
  
void float_image_get_interp_indices_and_weights
  ( double z, 
    int N, 
    int *i0, 
    int *i1, 
    double *w0,
    double *w1
  )
  {
    /* Get the `low' index {i0}: */
    (*i0) = (int)floor(z - 0.5);
    
    /* Compute the interpolation weights {w0} and {w1}: */
    (*w1) = z - (*i0) - 0.5;
    if ((*w1) < 0.0) { (*i0)--; (*w1) += 1.0; }
    if ((*w1) > 1.0) { (*i0)++; (*w1) -= 1.0; }
    (*w0) = 1.0 - (*w1);
   
    /* Compute the superior index {i1}: */
    (*i1) = (*i0) + 1;

    /* Clip {i0} and {i1} to {0..N-1}: */
    if ((*i0) < 0) 
      { (*i0) = (*i1) = 0; }
    else if ((*i1) >= N)
      { (*i0) = (*i1) = N-1; }
  }

void float_image_update_sample_range(float_image_t *fim, int c, float *vmin, float *vmax)
  { int NC = fim->sz[0];
    int NX = fim->sz[1];
    int NY = fim->sz[2];
    if ((c < 0) || (c >= NC))
      { /* Invalid channel, all samples are zero. (But there may be no samples!) */
        if ((NX > 0) && (NY > 0))
          { float v = 0.0;
            if (v > (*vmax)) { (*vmax) = v; }
            if (v < (*vmin)) { (*vmin) = v; }
          }
      }
    else
      { /* Valid channel. */
        int x, y;
        for (y = 0; y < NY; y++)
          { for (x = 0; x < NX; x++)
              { float v = float_image_get_sample(fim, c, x, y);
                if (v > (*vmax)) { (*vmax) = v; }
                if (v < (*vmin)) { (*vmin) = v; }
              }
          }
      }
  }

#define float_image_max_samples (1024u*1024u*1024u)
  /* Maximum total elements (2^30), for sanity checks. */

float_image_t *float_image_new (int NC, int NX, int NY)
  { /* Sanity checks: */
    demand((NC >= 0) && (NC < float_image_max_size), "too many channels");
    demand((NX >= 0) && (NX < float_image_max_size), "too many columns");
    demand((NY >= 0) && (NY < float_image_max_size), "too many rows");
    ix_pos_count_t NS = ((ix_pos_count_t)NC)*NX*NY;
    demand(NS < float_image_max_samples, "too many samples");
    /* Allocate header: */
    float_image_t *fim = (float_image_t *)notnull(malloc(sizeof(float_image_t)), "no mem");
    fim->sz[0] = NC; fim->st[0] = (NC < 2 ? 0 : 1);
    fim->sz[1] = NX; fim->st[1] = (NX < 2 ? 0 : NC);
    fim->sz[2] = NY; fim->st[2] = (NY < 2 ? 0 : NC*NX);
    fim->bp = 0;
    if ((NC == 0) || (NX == 0) || (NY == 0))
      { fim->sample = (float *)NULL; }
    else
      { fim->sample = (float *)notnull(malloc(NS*sizeof(float)), "no mem"); }
    (void)ix_parms_are_valid(3, fim->sz, fim->bp, fim->st, /*die:*/ TRUE);
    return fim;
  }
  
void float_image_free(float_image_t *fim)
  { if (fim == NULL) return;
    if (fim->sample != NULL) { free(fim->sample); }
    free(fim);
  }
  
void float_image_get_size(float_image_t *IMG, int *NC, int *NX, int *NY)
  { (*NC) = IMG->sz[0]; (*NX) = IMG->sz[1]; (*NY) = IMG->sz[2]; }
  
void float_image_check_size(float_image_t *IMG, int NC, int NX, int NY)
  { demand(IMG->sz[0] == NC, "wrong number of channels");
    demand(IMG->sz[1] == NX, "wrong number of columns");
    demand(IMG->sz[2] == NY, "wrong number of rows");
  }

#define float_image_file_version "2006-03-25"

void float_image_write(FILE *wr, float_image_t *fim)
  { 
    int NC = fim->sz[0];
    int NX = fim->sz[1];
    int NY = fim->sz[2];
    filefmt_write_header(wr, "float_image_t", float_image_file_version);
    fprintf(wr, "NC = %d\n", NC);
    fprintf(wr, "NX = %d\n", NX);
    fprintf(wr, "NY = %d\n", NY);
    int c, x, y;
    for(y = 0; y < NY; y++)
      { if (y > 0) { fprintf(wr, "\n"); }
        for(x = 0; x < NX; x++)
          { fprintf(wr, "%5d %5d", x, y);
            for (c = 0; c < NC; c++) 
              { float v = float_image_get_sample(fim, c, x, y);
                fprintf(wr, " %+14.7e", v);
              }
            fprintf(wr, "\n");
          }
      }
    filefmt_write_footer(wr, "float_image_t");
    fflush(wr);
  }

float_image_t *float_image_read(FILE *rd)
  {
    filefmt_read_header(rd, "float_image_t", float_image_file_version);
    int NC = nget_int(rd, "NC"); fget_eol(rd);
    int NX = nget_int(rd, "NX"); fget_eol(rd);
    int NY = nget_int(rd, "NY"); fget_eol(rd);
    float_image_t *fim = float_image_new(NC, NX, NY);
    int c, x, y;
    for (y = 0; y < NY; y++)
      { if (y > 0) { fget_eol(rd); }
        for (x = 0; x < NX; x++)
          { int xr = fget_int(rd);
            int yr = fget_int(rd);
            demand((xr == x) && (yr == y), "bad pixel indices");
            /* Read channels of pixel {(x,y)}: */
            for (c = 0; c < NC; c++)
              { double v = fget_double(rd);
                demand(fabs(v) <= MAXFLOAT, "bad sample value");
                float_image_set_sample(fim, c, x, y, v);
              }
            fget_eol(rd);
          }
      }
    filefmt_read_footer(rd, "float_image_t");
    return fim;
  }

