/* See {float_image_align.h}. */
/* Last edited on 2007-09-06 13:51:40 by stolfi */

#include <assert.h>
#include <values.h>
#include <string.h>
#include <math.h>
 
#include <bool.h>
#include <r2.h>
#include <i2.h>
#include <jsmath.h>
#include <affirm.h>
#include <float_image.h>
#include <float_image_align.h>

/* INTERNAL PROTOTYPES */

int fial_ix_first(int vmin, int vmax);
  /* Returns the first value in the sequence {SEQ =
    (0,-1,+1,-2,+2,..)} that lies in the range {vmax..vmin}. */

int fial_ix_next(int v, int vmin, int vmax);
  /* Returns the next integer after {v} in the sequence {SEQ} that
    lies in the range {vmax..vmin}.  If there is no such integer,
    returns some integer outside that range. */

bool_t fial_next_tuple(int n, i2_t d[], i2_t ns);
  /* Given a step count tuple {d[0..n-1]} of integer displacements,
    replaces it by the next step count tuple in lexicographic order,
    among the almost-balanced tuples whose coordinates along axis {k}
    lie in the range {-ns.c[k]..+ns.c[k]}. */

void fial_displace_points(int n, r2_t p[], i2_t d[], r2_t q[], r2_t step);
  /* Sets {q[0..n-1]} to the points {p[0..n-1]} displaced by the step counts
    {d[0..n-1]} times the given step lengths {step}. 

    More precisely sets {q[i].c[k] = p[i].c[k] + d[i].c[k]*step.c[k]}, 
    for {i} in {0..n-1} and {k} in {0,1}. */
    
void fial_debug_displacement(char *head, int n, i2_t d[], r2_t q[], double Q, char *tail);
  /* Prints the displacement step counts {d[0..n-1]}, the displaced
    points {q[0..n-1]}, and the mismatch value {Q} to {stderr},
    prefixed by {head} and siffixed by {tail}. */

/* IMPLEMENTATIONS */

void fial_optimal_shifts
  ( int n,
    fial_mismatch_t *mm, /* Funtion that evaluates the mismatch between the images. */
    r2_t step,           /* Precision required along each axis. */
    i2_t ns,             /* Max number of adjustment steps along each axis. */
    r2_t p[],            /* (IN/OUT) Corresponding points in each image. */
    double *QP           /* (OUT) Discrepancy measure for best alignment. */
  )
  { 
    bool_t debug = TRUE;
    
    int i, k;
      
    /* Check search grid parameters: */
    for (k = 0; k < 2; k++)
      { demand(ns.c[k] >= 0, "search step count must be non-negative");
        if (ns.c[k] > 0)
          { demand(step.c[k] > 0, "search {step} must be positive"); }
      }
    
    i2_t d[n];          /* Adjustment step counts for each coordinate. */
    i2_t dbest[n];      /* Best step count tuple found so far. */
    double Qbest = INF; /* Mismatch of best alignment. */

    r2_t q[n];      /* Work: Displaced points {q[i] = p[i] + d[i]}. */
        
    /* Initialize the adjustment step counts {d[0..n-1]}: */
    for (i = 0; i < n; i++) { d[i] = (i2_t){{0,0}}; }
    
    /* Enumeration loop: */
    while (1) 
      { /* Evaluate {Q} for the adjustment step count tuple {d[0..n-1]}: */
        fial_displace_points(n, p, d, q, step); 
        double Q = mm(n, q);
        if (debug) { fial_debug_displacement("trial", n, d, q, Q, "\n"); }
        if (Q < Qbest) { Qbest = Q; for (i = 0; i < n; i++) { dbest[i] = d[i]; } } 
        /* Advance to the next tuple: */
        if (fial_next_tuple(n, d, ns)) 
          { /* Enumeration finished, return the best solution: */
            fial_displace_points(n, p, dbest, p, step); 
            (*QP) = Qbest;
            return;
          }
      }
  }    

void fial_displace_points(int n, r2_t p[], i2_t d[], r2_t q[], r2_t step)
  { int i, k;
    for (i = 0; i < n; i++)
      { for (k = 0; k < 2; k++)
          { q[i].c[k] = p[i].c[k] + d[i].c[k]*step.c[k]; }
      }
  }


bool_t fial_next_tuple(int n, i2_t d[], i2_t ns)
  {
    /* There are {2*n} coordinates to be varied, each identified by
      an image index {i} and an axis index {k}.  For brevity, 
      we will write {ik} for {2*i + k}, and `dc[ik]' for {d[i].c[k]}. */

    int i;  /* Image index */
    int k;  /* Axis index. */
    
    /* Compute sum {s} of {d[0..n-1]}: */
    i2_t s = (i2_t){{0,0}}; /* Sum of displacements along each axis. */
    for (i = 0; i < n; i++)
      { for (k = 0; k < n; k++)
          { s.c[k] += d[i].c[k]; }
      }
      
    /* Ranges for the sum {s} along either axis: */
    int smin = -n/2, smax = smin + n - 1; 
    
    /* Find the first coord {dc[ik]} in {dc[0..2*n-1]} that can be bumped: */
    i = 0; k = 0;
    while (TRUE)
      { /* Now {s} is the sum of all coords {dc[ik..2*n-1]} */
        if (i >= n)
          { /* Exhausted all possibilities, stop: */
            return TRUE;
          }
        /* Update {s} to the sum of coords {dc[ik+1..2*n-1]}: */
        s.c[k] -= d[i].c[k];
        /* Find the range {[vmin..vmax]} for coord {dc[ik]}: */ 
        int vmin = imax(-ns.c[k], smin - s.c[k] - i*ns.c[k]);  /* Min valid value. */
        int vmax = imin(+ns.c[k], smax - s.c[k] + i*ns.c[k]);  /* Max valid value. */
        /* Find the next valid value {v} of coord {dc[ik]} in sequence {SEQ}: */
        int v = fial_ix_next(d[i].c[k], vmin, vmax);
        if ((v >= vmin) && (v <= vmax)) 
          { /* Bump coord {ik} to {v}, and update sum {s}: */
            d[i].c[k] = v;
            s.c[k] += v;
            break;
          }
        else 
          { /* This coord can'i be incremented, carry to the next one: */
            int ik = (2*i + k) + 1; i = ik / 2; k = ik % 2; 
          }
      }
    /* Here {s} is the sum of coords {dc[ik..2*n-1]}. */
    /* Reset coords {dc[0..ik-1]} to the their earliest possible values: */
    while (2*i + k > 0)
      { /* Move {ik} down: */
        int ik = (2*i + k) - 1; i = ik / 2; k = ik % 2; 
        /* Here {s} is the sum of digits {ik+1..2*n-1}. */
        /* Find range {[vmin..vmax]} for digit {ik}: */ 
        int vmin = imax(-ns.c[k], smin - s.c[k] - i*ns.c[k]);  /* Min valid value. */
        int vmax = imin(+ns.c[k], smax - s.c[k] + i*ns.c[k]);  /* Max valid value. */
        /* Pick first valid value {v} for digit {ik} in {0,-1,+1,-2,+2,...-ns[k],+ns[k]}: */
        int v = fial_ix_first(vmin, vmax);
        assert((v >= -ns.c[k]) && (v <= +ns.c[k]));
        /* Set digit {ik} to {v} and update the sum {s.c[k]}: */
        d[i].c[k] = v;
        s.c[k] += v;
      }     
    assert((s.c[0] >= smin) && (s.c[0] <= smax)); 
    assert((s.c[1] >= smin) && (s.c[1] <= smax)); 
    return FALSE;
  }

int fial_ix_first(int vmin, int vmax)
  {
    if (vmin > 0)
      { return vmin; }
    else if (vmax < 0)
      { return vmax; }
    else
      { return 0; }
  }

int fial_ix_next(int v, int vmin, int vmax)
  {
    if (vmin > 0)
      { return (v < vmin ? vmin : v + 1); }
    else if (vmax < 0)
      { return (v > vmax ? vmax : v - 1); }
    else
      { /* Range {vmin..vmax} contains zero. Try the next value from {SEQ}: */
        if (v < 0)
          { v = -v; }
        else 
          { v = -(v + 1); }
        if ((v >= vmin) && (v <= vmax)) { return v; }
        /* Try again with the next value: */
        if (v < 0)
          { v = -v; }
        else 
          { v = -(v + 1); }
        if ((v >= vmin) && (v <= vmax)) { return v; }
        /* Two successive failures; it means exhaustion: */
        return (vmax > -vmin ? 1 + vmax : 1 - vmin);
      }
  }

void fial_debug_displacement(char *head, int n, i2_t d[], r2_t q[], double Q, char *tail)
  {
    fprintf(stderr, "    %s\n", head);
    int i;
    i2_t s = (i2_t){{0,0}};
    for (i = 0; i < n; i++)
      { if (d != NULL) 
          { fprintf(stderr, "      d[%02d] = ( %+3d  %+3d )", i, d[i].c[0], d[i].c[1]); 
            s.c[0] += d[i].c[0];
            s.c[1] += d[i].c[1];
          }
        if (q != NULL) 
          { fprintf(stderr, "  q[%02d] = ( %+9.2f  %+9.2f )", i, q[i].c[0], q[i].c[1]); }
        fprintf(stderr, "\n");
      }
    if (d != NULL) 
      { fprintf(stderr, "      sum =   ( %+3d  %+3d )", s.c[0], s.c[1]); }
    fprintf(stderr, "    Q = %12.5f\n", Q);
    fprintf(stderr, "%s", tail);
  }

