#ifndef indexing_descr_H
#define indexing_descr_H

/* Multidimensional array descriptors. */
/* Last edited on 2007-10-10 20:51:12 by stolfi */ 

#include <stdio.h>
#include <bool.h>
#include <indexing.h>

/* BASIC TYPES */

/* NUMBER OF AXES IN DESCRIPTOR */

#define ix_descr_NAXES 6
  /* Number of indices (dimensions, axes) in any array descriptor. Lower-dimensional
    arrays are obtained by setting the size to 1 along the unwanted axes. */

#define NA ix_descr_NAXES
  /* A shorter local name for {ix_descr_NAXES}. */

/* DESCRIPTORS */

typedef struct ix_descr_t 
  { ix_size_t sz[NA];  /* Extent of array's domain along each axis. */
    ix_pos_t bp;       /* Position of element {[0,..0]}. */
    ix_step_t st[NA];  /* Position increment for each index. */
  } ix_descr_t; 
  /* An indexing formula for arrays with {NA} indices.
    The position of an element with index tuple {ix[0..NA-1]} is 
    {D.bp + ix[0]*D.st[0] + ix[NA-1]*D.st[NA-1]]}.  See {indexing.h}
    for more details. */

/* STANDARD STEP COMPUTATION */

ix_descr_t ix_descr_from_sizes ( ix_size_t sz[] );
  /* Returns a descrptor record {D} for an array that has size {sz[i]}
    along axis {i}, for {i} in {0..NA-1}. The base {bp} and the
    increments {st[0..NA-1]} are computed so as to imply a compact set
    of element positions, statring from 0, in the Fortran order
    (first index has increment 1).
    
    Specifically, sets the sizes {D.sz[0..NA-1]} to
    {sz[0..NA-1]}, the base pointer {D.bp} to 0,
    the increment {D.st[0]} to 1, and increment {D.st[i]}
    to {D.st[i-1]*sz[i-1]}, for {i} in {1..NA-1}. */

/* ACESSING INDIVIDUAL ELEMENTS */

ix_pos_t ix_descr_position ( ix_descr_t *D, ix_index_t ix[] );
  /* Returns the position of the element of {D} with indices {ix[0..NA-1]}. */

/* INDEX TUPLE MANIPULATION */

void ix_descr_indices_fill ( ix_index_t ix[], ix_index_t val );
  /* Sets {ix[i] = val} for {i=0,..NA-1}. */

void ix_descr_indices_assign ( ix_index_t ix[], ix_index_t val[] );
  /* Sets {ix[i] = val[i]} for {i=0,..NA-1}. */

void ix_descr_indices_shift ( ix_index_t ix[], ix_index_t inc[] );
  /* Sets {ix[i] += inc[i]} for {i=0,..NA-1}. */

bool_t ix_descr_indices_first ( ix_descr_t *D, ix_index_t ix[] );
  /* Sets {ix[i] = 0} for {i=0,..NA-1}, and returns TRUE iff that
    is a valid index tuple for {D} (i.e. {D.sz[i]>0} for all {i}). */

bool_t ix_descr_indices_last ( ix_descr_t *D, ix_index_t ix[] );
  /* Sets {ix[i] = D.sz[i]-1} for {i=0,..NA-1}, and returns TRUE iff that
    is a valid index tuple (i.e. {D.sz[i]>0} for all {i}). */

bool_t ix_descr_indices_are_valid ( ix_descr_t *D, ix_index_t ix[] );
  /* Returns TRUE iff {ix[0..NA-1]} is a valid index tuple for {D};
    i.e. if {ix[i]} lies in the range {0..D.sz[i]-1}, 
    for {i=0,..NA-1}. */

/* SIZE TUPLE MANIPULATION */

void ix_descr_sizes_assign ( ix_size_t sza[], ix_size_t szb[] );
  /* Sets {sza[i] = szb[i]} for {i=0,..NA-1}. */

bool_t ix_descr_sizes_shrink ( ix_size_t sza[], ix_size_t szb[] );
  /* Sets {sza[i] = min(sza[i], szb[i])} for {i=0,..NA-1}.
    Returns TRUE iff any {sza[i]} was actually changed. */

bool_t ix_descr_sizes_expand ( ix_size_t sza[], ix_size_t szb[] );
  /* Sets {sza[i] = max(sza[i], szb[i])} for {i=0,..NA-1}.
    Returns TRUE iff any {sza[i]} was actually changed. */

ix_size_t ix_descr_max_size ( ix_descr_t *D );
  /* Returns the max dimension of {D} along any axis. */

ix_size_t ix_descr_min_size ( ix_descr_t *D );
  /* Returns the min dimension of {D} along any axis. */

/* ELEMENT COUNTING */

bool_t ix_descr_is_empty ( ix_descr_t *D );
  /* TRUE iff the array {D} is empty, i.e. {D.sz[i]==0} for some {i} in {0..NA-1}. */

ix_pos_count_t ix_descr_num_tuples ( ix_descr_t *D );
  /* Number of valid index tuples in the array {D}; 0 if the array is empty.
    Beware that virtual element replication may create huge arrays that exceed
    the capacity of an {ix_pos_count_t}. */

ix_pos_count_t ix_descr_num_positions ( ix_descr_t *D );
  /* Number of *distinct* element positions in {D}, not counting virtual replications.
    Assumes that the descriptor is valid (see {ix_descr_is_valid} below). */

ix_pos_t ix_descr_min_pos ( ix_descr_t *D );
ix_pos_t ix_descr_max_pos ( ix_descr_t *D );
  /* Minimum and maximum position of all elements of {D} with a valid
    index tuple. Undefined if the array is empty. */

bool_t ix_descr_same_size ( ix_descr_t *D, ix_descr_t *B, bool_t die );
  /* Returns TRUE if {D.sz[i] == B.sz[i]} for all {i} in {0..NA-1}, returns TRUE.
    Otherwise, returns FALSE (if {die == FALSE}) or fails with
    error (if {die == TRUE}). */ 

bool_t ix_descr_contained ( ix_descr_t *D, ix_descr_t *B, bool_t die );
  /* Returns TRUE if the domain of {D} is contained in the domain of {B} ---
    that is, if {D.sz[i] <= B.sz[i]} for all {i} in {0..NA-1}.
    Otherwise, returns FALSE (if {die == FALSE}) or fails with
    error (if {die == TRUE}). */ 

/* DESCRIPTOR MANIPULATION */

/* The following procedures modify the fields {D.sz}, {*D.bp} and
  {D.st} of the given descriptor {D} so as to change the set of valid
  indices and/or the mapping between indices and positions. Obviously,
  no array elements are actually allocated, reclaimed, or changed.
  
  In the describing comments, {ix'[0..NA-1]} is a generic tuple valid
  for {D} AFTER the call, referring to some position {p}; and
  {ix[0..NA-1]} is the tuple that, BEFORE the call, would yield the
  same position.
  
  All the procedures below preserve the validity of the descriptor
  {D} (as defined by {ix_descr_is_valid} below).  Unless said otherwise,
  the operations preserve the empty/non-empty status of the array. */

void ix_descr_crop ( ix_descr_t *D, ix_axis_t i, ix_size_t skip, ix_size_t keep );
  /* Crops the array described by {D} along axis {i} to the 
    index range {ini..fin}. 
  
    More precisely, sets {D.sz[i]} to {keep}, and modifies {D.bp,D.st}
    so that {ix[i] = ix'[i]+skip} and {ix[k] = ix'[k]} for all {k\neq i}.
    If {keep==0}, the array becomes empty (but retains its size along
    the other axes); in that case {D.bp} and {D.st[0..NA-1]} are reset to 0.
    The result is undefined if {skip+keep > D.sz[i]}.
    
    Thus, to clip a two-dimensional {m  n} matrix to its upper right
    quarter, use {ix_descr_crop(D, 0, 0,m/2)} and then
    {ix_descr_crop(D, 1, n/2,n-n/2)}, where {m=D.sz[0]} and {n=D.sz[1]}. 
    To reduce that matrix to its column number {c}, use
    {ix_descr_crop(D, 1, c,1)}. */

void ix_descr_subsample ( ix_descr_t *D, ix_axis_t i, ix_size_t stride );
  /* Subsamples the array described by {D} along axis {i} by taking
    one every {stride} slices. The sampling step {stride} must be positive. 
    
    More precisely, reduces {D.sz[i]} to {floor((D.sz[i]-1)/stride)+1}, and modifies
    {D.bp} so that {ix[i] = stride*ix'[i]} and {ix[k] = ix'[k]} for all {k\neq i}. */

void ix_descr_flip ( ix_descr_t *D, ix_axis_t i );
  /* Flips the array described by {D} along axis {i}. 
  
    More precisely, modifies {D.bp} and {D.st[i]} so that
    {ix[i] = D.sz[i]-1-ix'[i]}, and {ix[k] = ix'[k]} for all {k\neq i}. The
    size {D.sz[i]} is not changed. */

void ix_descr_replicate ( ix_descr_t *D, ix_axis_t i, ix_size_t size );
  /* Virtually replicates the array described by {D} along axis {i}, {size} times.
  
    On input, the array must be trivial along axis {i} (i.e., {D.sz[i]} must be 1).
    The procedure sets {D.sz[i]=size} and {D.st[i]=0}, so that {ix[i]=0} and
    {ix[k]=ix'[k]} for all {k\neq i}. The replication count {size} must be
    positive. */

void ix_descr_swap_indices ( ix_descr_t *D, ix_axis_t i, ix_axis_t j );
  /* Transposes the array described by {D}, by exchanging the axes {i} and {j}. 
  
    More precisely, swaps {D.sz[i]} with {D.sz[j]} and {D.st[i]} with
    {D.st[j]}, so that {ix[i]=ix'[j]}, {ix[j] = ix'[i]}, and
    {ix[k]=ix'[k]} for all {k\neq i,j}. If {i == j}, the procedure has
    no effect.
    
    Thus, to transpose a two-dimensional matrix, in the ordinary
    sense, use {ix_descr_swap_indices(D, 0,1)}. To reduce it to
    its row number {r}, viewed as a column matrix, use
    {ix_descr_crop(D, 0, r,1)} and then
    {ix_descr_swap_indices(D, 0,1)}. */

void ix_descr_flip_indices ( ix_descr_t *D, ix_axis_t i, ix_axis_t j );
  /* Reverses the order of all indices of array described by {D} 
    between axis {i} and axis {j}, inclusive.
  
    More precisely: if {i<j}, the procedure swaps {D.sz[i+k]} with
    {D.sz[j-k]} and {D.st[i+k]} with {D.st[j-k]} for all {k} in
    {0..(j-i-1)/2}; thus {ix[i+k] = ix'[j-k]} for all those {k}. 
    It is a no-op if {i>=j}. */

void ix_descr_diagonal ( ix_descr_t *D, ix_axis_t i, ix_axis_t j );
  /* Extracts a diagonal slice of the array described by {D}, parallel to the
    bisector of axes {i} and {j}; and redefines axis {i} to be along
    that diagonal. Thus, using index values {r} and {s} for those axes,
    after the call, is equivalent to using {r} and {r+s} before the call.
  
    Upon entry, requires {i != j} and {0 < D.sz[i] <= D.sz[j]}. The
    procedure reduces {D.sz[j]} by {D.sz[i]-1}, and adds {D.st[j]} to
    {D.st[i]}; so that {ix[i]=ix'[i]}, {ix[j]=ix'[j]+ix'[i]}, and
    {ix[k]=ix'[k]} for all other axes {k}. If the array was non-empty,
    it will remain so.
    
    Thus, for example, if {D} is a {1001205} matrix, the call
    {ix_descr_diagonal(D,0,1)} will chop it down to a diagonal
    band, starting with the main diagonal and extending 21 elements to
    the right, indexed as a {100215} matrix. */

void ix_descr_chop ( ix_descr_t *D, ix_axis_t i, ix_size_t stride, ix_axis_t j );
  /* Chops the array described by {D} perpendicularly to axis {i} into
    fat slices of thickness {stride}, that are stacked along axis {j}. 
  
    Upon entry, the array size {D.sz[i]} along axis {i} must be a
    multiple of {stride}, and the array must be trivial along axis {j}
    ({D.sz[j] == 1}). The procedure sets {D.sz[j] = D.sz[i]/stride},
    {D.sz[i] = stride}, and modifies {st} so that {ix[i] = (ix'[i] +
    stride*ix'[j])}, {ix[j] = 0}. The effect is undefined if {stride
    == 0} or {j == i}
    
    For exmple, {ix_descr_chop(D, 0, 100, 1)} will turn a vector of
    5000 elements (along axis 0) into a two-dimensional arryay with
    100 elements per column (along index 0) and 50 columns (along axis 1). */

/* INDEX TUPLE STEPPING */

bool_t ix_descr_next ( ix_descr_t *D, ix_index_t ix[], ix_pos_t *p );
  /* Increments the index tuple {ix} to the next valid tuple in the
    Fortran indexing order (i.e., with the *first* index in the
    innermost loop).
    
    More precisely, the procedure tries to
    increment {ix[0]} first. If an index {ix[i]} to be incremented has
    reached its limit {D.sz[i]-1}, resets all indices {ix[0..i]} to 0,
    and tries to increment index {ix[i+1]} instead. If this `carry'
    process falls off the index -- that is, if the tuple {ix} changes
    from {D.sz[0..NA-1] - (1,..1)} back to {(0,..0)} -- the procedure
    returns TRUE; otherwise it returns FALSE. The effect is undefined
    if any size {D.sz[i]} is 0, or any index {ix[i]} is not in {0..D.sz[i]-1}.
    
    If {p} is not null, assumes that {*p} is the position corresponding
    to the index tuple {ix}, and adjusts it to account for the change.
  
    This procedure can be used to vary the index tuple {ix[0..NA-1]}
    over all the elements of an array with descriptor {D}, in a single
    loop, e.g.
    
      | if (ix_descr_assign_min(D,ix))
      |   { ix_pos_t p = ix_descr_position(D,ix);
      |     do { ... } while (! ix_descr_next(D,ix,&p));
      |   }

    */
    
bool_t ix_descr_prev ( ix_descr_t *D, ix_index_t ix[], ix_pos_t *p );
  /* Like {ix_descr_next}, but in the reverse order. 
    
    Namely, tries to decrement each index {ix[i]}, starting with
    {ix[0]}. If the index is already zero, resets all indices
    {ix[0..i]} to their maximum values {D.sz[0]-1,.. D.sz[i]-1}. and tries
    again with the next index {ix[i+1]}. 
    
    The returned result is TRUE iff this `borrow' falls off the index
    tuple -- that is, {ix} changes from {(0,..0)} back to {D.sz[0..NA-1]
    - (1,..1)}.  The position {*p} is updated if {p} is not NULL.
    
    Thus, to scan an array in reverse order, use
    
      | if (ix_descr_assign_max(D,ix))
      |   { ix_pos_t p = ix_descr_position(D,ix);
      |     do { ... } while (! ix_descr_prev(D,ix,&p));
      |   }

    */

sign_t ix_descr_compare ( ix_index_t ixa[], ix_index_t ixb[] );
  /* These procedures return -1, 0, or +1 depending on whether the
    index tuple {ixa} comes before, coincides, or follows {ixb} in the
    standard (Fortran-like) scanning order. The result is undefined if
    any index is negative. */
    
/* ELEMENT ENUMERATION */

typedef void ix_descr_op_t ( ix_index_t ix[], ix_pos_t pA, ix_pos_t pB, ix_pos_t pC );
  /* Client procedure for {ix_descr_enum}. */

void ix_descr_enum ( ix_descr_op_t op, ix_descr_t *A, ix_descr_t *B, ix_descr_t *C );
  /* Applies the operation {op} to all valid index tuples 
    (and corresponding positions) of three arrays {A,B,C} in parallel.
    
    More precisely, let {sz[i]} be the minimum of
    {A.sz[i],B.sz[i],C.sz[i]}, for {i} in {0..NA-1}. The procedure
    enumerates all valid index tuples {ix[0..NA-1]} between {(0,..0)}
    and {sz-(1,..1)}. For each tuple {ix}, it computes the the corresponding positions
    {pA,pB,pC} in {A,B,C}, respectively; and calls {op(ix,pA,pB,pC)}.
    
    If {A} is NULL, {A.sz[i]} is assumed to be {+oo}, and {pA} is
    always 0. Ditto for the other two descriptors. If all three
    descriptors are NULL, the procedure is a no-op.
    
    The index tuples are enumerated in the C/Pascal order (with the
    last index in the innermost loop), and each index {ix[i]} increases
    from 0 to {sz[i]-1}. Use {ix_descr_flip_indices} to get the FORTRAN
    order, and {ix_descr_flip} to scan some or all indices in reverse.
    
    Note that virtually replicated array elements will be visited
    multiple times (once for each valid index tuple that refers to
    them). A repeated visit to an element of {A} can be recognized by
    {(ix[i]>0) && (A != NULL) && (A.st[i]==0)} for some {i}; and ditto for
    the other two descriptors. */

/* PARAMETER VALIDATION */

bool_t ix_descr_is_valid ( ix_descr_t *D, bool_t die );
  /* Checks whether the descriptor {D} satisfies the following conditions:

      (0) Every valid index tuple yields a non-negative  position.
        
      (1) For any {i}, if {D.sz[i]==1}, then {D.st[i]==0}.

      (2) If any {D.sz[i]} is zero, then {D.bp} and all 
        increments {D.st[0..NA-1]} are zero.

    The procedure returns TRUE if these conditions hold. Otherwise,
    it either aborts the program with an error message (if {die=TRUE})
    or returns FALSE (if {die=FALSE}). */

bool_t ix_descr_positions_are_distinct ( ix_descr_t *D, bool_t die );
  /* Checks whether the indexing parameters of {D} imply distinct
    positions for distinct valid index tuples, except for virtual
    replication. That is, whether two tuples {ixa[0..NA-1]} and
    {ixb[NA-1]} give the same position iff and only iff
    {ixa[i]!=ixb[i]} implies {D.st[i]=0}.
    
    The procedure assumes that the descriptor valid in the sense
    of {ix_descr_is_valid}. It returns TRUE if the distinctness
    condition is satisfied; otherwise, it either aborts the program
    with an error message (if {die=TRUE}) or returns FALSE (if
    {die=FALSE}). */

/* INPUT/OUTPUT */

void ix_descr_print_descr ( FILE *wr, char *pre, ix_descr_t *D, int wd, char *suf );
  /* Writes the descriptor {D} to {wr}, in readable ASCII format. 
    The output consists of four lines 
    
      | "axes = {d}" 
      | "base = {D.bp}"
      | "step = {D.st[0]} {D.st[1]} ... {D.st[d-1]}"
      | "size = {D.sz[0]} {D.sz[1]} ... {D.sz[d-1]}"
    
    where {d} is the largest integer such that {D.sz[d-1] != 1}.
    
    Each line is prefixed by the string {pre} (which defaults to "" if NULL) 
    and suffixed with the string {suf} (which defaults to "\n" if NULL).  Each size or
    step is padded on the left to total width {wd} or more. Elements are separaed by 
    blanks. Positive {step} elements will have an explicit "+" sign. */

/* NAMESPACE CLEANUP */

#undef NA
  /* External clients should use {ix_descr_NAXES}. */

#endif
