/* Tests the primitive splines */
/* Last edited on 2007-10-15 00:29:37 by stolfi */

#define _GNU_SOURCE
#include <dg_grid.h>
#include <dg_pulse.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define N 30
  /* Number of samples per interval. */

#define MAXC 2
  /* Maximum continuity order. */
  
#define MAXDER 2
  /* Maximum derivative order to evaluate. */
  
#define BIGSZ 64
  /* A power of 2 greater than the support size of any spline. */

#define dx 0.0001
  /* Step for numerical derivation. */

#define delta ((MAXDER+1)*dx)
  /* Min distance to keep from knots. Must be greater than {MAXDER*dx}.  */
  
#define DIR "plt"
  /* Directory for output files. */

/* INTERNAL PROTOS */

int main (int argc, char **argv);

void show_family ( dg_pulse_family_t *fam );

void show_pulse ( dg_pulse_family_t *fam, dg_pulse_t *p );

char *tsp_make_name
  ( dg_pulse_family_t *fam,
    dg_pulse_t *p
  );
    
void sample
  ( FILE *wr, 
    dg_pulse_family_t *fam,
    dg_pulse_t *p,
    double x, 
    dg_cont_t ord,
    double e[]
  );

/* IMPLEMENTATIONS */

int main (int argc, char **argv)
  { int c;
    dg_pulse_kind_t pkind;
    for (pkind = 0; pkind <= dg_pulse_kind_MAX; pkind++)
      { for (c = -1; c <= MAXC; c++)
          { int maxg = 2*c + 3;
            int g;
            for (g = 0; g <= maxg; g++)
              { dg_pulse_family_t fam = dg_pulse_family(c, g, pkind);
                show_family(&fam);
              }
          }
      }
    fprintf(stderr, "done.\n");
    return 0;
  }
                
void sample
  ( FILE *wr, 
    dg_pulse_family_t *fam, 
    dg_pulse_t *p, 
    double x, 
    dg_cont_t ord,
    double e[]
  )
  {
    int k, j;
    interval_t R = (interval_t){{ 0.0, (double)p->gsz }};
    
    /* Get pulse value and analyic derivatives {f[0..MAXDER]}: */
    double f[ord+1];
    for (k = 0; k <= ord; k++) { f[k] = 0; }
    dg_pulse_eval(fam, p, &R, x, ord, f);
    
    /* Get pulse value and numerical derivatives {d[0..MAXDER]}: */
    double d[ord+1];
    for (k = 0; k <= ord; k++)
      { dg_pulse_eval(fam, p, &R, x + k*dx, 0, &(d[k])); }
    for (k = 1; k <= ord; k++)
      { for (j = ord; j >= k; j--) { d[j] = (d[j]-d[j-1])/dx; } }

    /* Print values and derivs, and update the max error {e}. */
    fprintf(wr, "%8.5f", x);
    for (k = 0; k <= ord; k++) 
      { fprintf(wr, "  %10.5f %10.5f", f[k], d[k]);
        double err = fabs(f[k]-d[k]);
        if (err > e[k]) { e[k] = err; }
      }
    for (k = ord+1; k<= MAXDER; k++)
      { fprintf(wr, "  %10s %10s", "0", "0"); }
    fprintf(wr, "\n");
  }

void show_family ( dg_pulse_family_t *fam )
  {
    dg_pulse_mother_index_t pix;
    for (pix = 0; pix < fam->nmp; pix++)
      { dg_grid_size_t msz = dg_pulse_mother_supp_count(fam, pix);
        dg_grid_size_t gsz = 1;
        do
          { dg_pulse_t p = dg_pulse(fam, pix, gsz, 0); 
            show_pulse(fam, &p);
            gsz = 2*gsz;
          }
        while (gsz <= 2*msz);
      }
  }
  
void show_pulse ( dg_pulse_family_t *fam, dg_pulse_t *p )
  {
    int iz;
    char *fname = tsp_make_name(fam, p);
    FILE *wr;
    /* Max derivation order to ask: */
    dg_cont_t ord = fam->g+1; 
    if (ord > MAXDER) { ord = MAXDER; }
    /* Max numerical-analytical discrepancy in each derivative: */
    double e[ord+1];
    { int k; for (k = 0; k <= ord; k++) { e[k] = 0; } }
    wr = open_write(fname, TRUE);
    for (iz = -2; iz <= (p->msz)*N+2; iz++)
      { double x = ((double)iz)/((double)N);
        if ((iz % N) == 0)
          { sample(wr, fam, p, x-delta, ord, e);
            sample(wr, fam, p, x+delta, ord, e);
          }
        else
          { sample(wr, fam, p, x, ord, e); }
      }
    if (wr == stdout) { fflush(wr); } else { fclose(wr); }
    fprintf(stderr, "max errors in derivatives: \n");
    { int k;
      for (k = 0; k <= MAXDER; k++) 
        { fprintf(stderr, " %24.16e", e[k]); }
    }
    fprintf(stderr, "\n");
  }

char *tsp_make_name
  ( dg_pulse_family_t *fam,
    dg_pulse_t *p
  )
  {
    char *fname = NULL; 
    char pkind_txt = dg_pulse_kind_to_char(fam->pkind);
    char *c_txt = NULL;
    if (fam->c >= 0) { asprintf(&c_txt, "%d", fam->c); } else { c_txt = "-"; }
    char *gsz_txt = NULL;
    if (p->gsz < p->msz) { asprintf(&gsz_txt, "%02lld", p->gsz); } else { gsz_txt = "--"; }
    asprintf
      ( &fname, 
        "%s/pu-z%s-%c-c%s-g%02d-m%02d.plt", 
        DIR, gsz_txt, pkind_txt, c_txt, fam->g, p->pix
      );
    if (fam->c >= 0) { free(c_txt); }
    return fname;
  }
